@@ -19,15 +19,17 @@ module should be used only with the library version it accompanied.
 
 import sys
 from ctypes import *
-from binascii import hexlify, unhexlify
+from binascii import hexlify, unhexlify, b2a_base64
+from datetime import datetime
+import time
 
 """
 Module initialization
 """
 if sys.platform == 'darwin':
-    botan = CDLL('libbotan-1.11.dylib')
+    botan = CDLL('libbotan-2.dylib')
 else:
-    botan = CDLL('libbotan-1.11.so')
+    botan = CDLL('libbotan-2.so')
 
 expected_api_rev = 20151015
 botan_api_rev = botan.botan_ffi_api_version()
@@ -308,16 +310,16 @@ def bcrypt(passwd, rng, work_factor = 10):
     out_len = c_size_t(64)
     out = create_string_buffer(out_len.value)
     flags = c_uint32(0)
-    rc = botan.botan_bcrypt_generate(out, byref(out_len), passwd, rng.rng, c_size_t(work_factor), flags)
+    rc = botan.botan_bcrypt_generate(out, byref(out_len), _ctype_str(passwd), rng.rng, c_size_t(work_factor), flags)
     if rc != 0:
         raise Exception('botan bcrypt failed, error %s' % (rc))
-    b = out.raw[0:out_len.value]
+    b = out.raw[0:out_len.value-1]
     if b[-1] == '\x00':
         b = b[:-1]
     return b
 
 def check_bcrypt(passwd, bcrypt):
-    rc = botan.botan_bcrypt_is_valid(passwd, bcrypt)
+    rc = botan.botan_bcrypt_is_valid(_ctype_str(passwd), bcrypt)
     return (rc == 0)
 
 """
@@ -340,11 +342,11 @@ def pbkdf_timed(algo, password, out_len, ms_to_run = 300, salt = rng().get(12)):
 """
 KDF
 """
-def kdf(algo, secret, out_len, salt):
-    botan.botan_kdf.argtypes = [c_char_p, POINTER(c_char), c_size_t, POINTER(c_char), c_size_t, POINTER(c_char), c_size_t]
+def kdf(algo, secret, out_len, salt, label):
+    botan.botan_kdf.argtypes = [c_char_p, POINTER(c_char), c_size_t, POINTER(c_char), c_size_t, POINTER(c_char), c_size_t, POINTER(c_char), c_size_t]
     out_buf = create_string_buffer(out_len)
     out_sz = c_size_t(out_len)
-    botan.botan_kdf(_ctype_str(algo), out_buf, out_sz, secret, len(secret), salt, len(salt))
+    botan.botan_kdf(_ctype_str(algo), out_buf, out_sz, secret, len(secret), salt, len(salt), label, len(label))
     return out_buf.raw[0:out_sz.value]
 
 """
@@ -595,25 +597,54 @@ class pk_op_key_agreement(object):
                                       lambda b,bl: botan.botan_pk_op_key_agreement(self.op, b, bl,
                                                                                    other, len(other),
                                                                                    salt, len(salt)))
-
+"""
+X.509 certificates
+"""
 class x509_cert(object):
-    def __init__(self, filename):
-        botan.botan_x509_cert_load_file.argtypes = [POINTER(c_void_p), c_char_p]
-        self.x509_cert = c_void_p(0)
-        botan.botan_x509_cert_load_file(byref(self.x509_cert), _ctype_str(filename))
+    def __init__(self, filename=None, buf=None):
+        if filename is None and buf is None:
+            raise ArgumentError("No filename or buf given")
+        if filename is not None and buf is not None:
+            raise ArgumentError("Both filename and buf given")
+        elif filename is not None:
+            botan.botan_x509_cert_load_file.argtypes = [POINTER(c_void_p), c_char_p]
+            self.x509_cert = c_void_p(0)
+            botan.botan_x509_cert_load_file(byref(self.x509_cert), _ctype_str(filename))
+        elif buf is not None:
+            botan.botan_x509_cert_load.argtypes = [POINTER(c_void_p), POINTER(c_char), c_size_t]
+            self.x509_cert = c_void_p(0)
+            botan.botan_x509_cert_load(byref(self.x509_cert), _ctype_bits(buf), len(buf))
 
     def __del__(self):
         botan.botan_x509_cert_destroy.argtypes = [c_void_p]
         botan.botan_x509_cert_destroy(self.x509_cert)
 
-    # TODO: have these convert to a python datetime
     def time_starts(self):
         botan.botan_x509_cert_get_time_starts.argtypes = [c_void_p, POINTER(c_char), POINTER(c_size_t)]
-        return _call_fn_returning_string(16, lambda b,bl: botan.botan_x509_cert_get_time_starts(self.x509_cert, b, bl))
+        starts = _call_fn_returning_string(16, lambda b,bl: botan.botan_x509_cert_get_time_starts(self.x509_cert, b, bl))
+        if len(starts) == 13:
+            # UTC time
+            struct_time = time.strptime(starts, "%y%m%d%H%M%SZ")
+        elif len(starts) == 15:
+            # Generalized time
+            struct_time = time.strptime(starts, "%Y%m%d%H%M%SZ")
+        else:
+            raise Exception("Wrong date/time format")
+
+        return datetime.fromtimestamp(time.mktime(struct_time))
 
     def time_expires(self):
         botan.botan_x509_cert_get_time_expires.argtypes = [c_void_p, POINTER(c_char), POINTER(c_size_t)]
-        return _call_fn_returning_string(16, lambda b,bl: botan.botan_x509_cert_get_time_expires(self.x509_cert, b, bl))
+        expires = _call_fn_returning_string(16, lambda b,bl: botan.botan_x509_cert_get_time_expires(self.x509_cert, b, bl))
+        if len(expires) == 13:
+            # UTC time
+            struct_time = time.strptime(expires, "%y%m%d%H%M%SZ")
+        elif len(expires) == 15:
+            # Generalized time
+            struct_time = time.strptime(expires, "%Y%m%d%H%M%SZ")
+        else:
+            raise Exception("Wrong date/time format")
+        return datetime.fromtimestamp(time.mktime(struct_time))
 
     def to_string(self):
         botan.botan_x509_cert_to_string.argtypes = [c_void_p, POINTER(c_char), POINTER(c_size_t)]
@@ -642,6 +673,17 @@ class x509_cert(object):
         botan.botan_x509_cert_get_public_key_bits.argtypes = [c_void_p, POINTER(c_char), POINTER(c_size_t)]
         return _call_fn_returning_vec(0, lambda b,bl: botan.botan_x509_cert_get_public_key_bits(self.x509_cert, b, bl))
 
+    def subject_public_key(self):
+        botan.botan_x509_cert_get_public_key.argtypes = [c_void_p, c_void_p]
+        pub = c_void_p(0)
+        botan.botan_x509_cert_get_public_key(self.x509_cert, byref(pub))
+
+        return public_key(pub)
+
+    def subject_dn(self, key, index):
+        botan.botan_x509_cert_get_subject_dn.argtypes = [c_void_p, c_char_p, c_size_t, POINTER(c_char), POINTER(c_size_t)]
+        return _call_fn_returning_string(0, lambda b,bl: botan.botan_x509_cert_get_subject_dn(self.x509_cert, _ctype_str(key), index, b, bl))
+
 
 """
 Tests and examples
@@ -657,7 +699,7 @@ def test():
     def test_kdf():
         print("KDF2(SHA-1)   %s" %
               hex_encode(kdf('KDF2(SHA-1)', hex_decode('701F3480DFE95F57941F804B1B2413EF'), 7,
-                             hex_decode('55A4E9DD5F4CA2EF82'))))
+                             hex_decode('55A4E9DD5F4CA2EF82'), hex_decode(''))))
 
     def test_pbkdf():
         print("PBKDF2(SHA-1) %s" %
@@ -673,6 +715,14 @@ def test():
         print('x %s' % hex_encode(psk))
         print('y %s\n' % (hex_encode(pbkdf('PBKDF2(SHA-256)', 'xyz', 32, iterations, salt)[2])))
 
+    def test_bcrypt():
+        print("Testing Bcrypt...")
+        r = rng()
+        phash = bcrypt('testing', r)
+        print("bcrypt returned %s (%d bytes)" % (hex_encode(phash), len(phash)))
+        print("validating the hash produced: %r" % (check_bcrypt('testing', phash)))
+        print("\n")
+
     def test_hmac():
 
         hmac = message_authentication_code('HMAC(SHA-256)')
@@ -719,7 +769,7 @@ def test():
             key = rng().get(kmax)
             pt = rng().get(21)
 
-            print("  plaintext %s (%d)"   % (hex_encode(pt), len(pt)))
+            print("  plaintext  %s (%d)"   % (hex_encode(pt), len(pt)))
 
             enc.set_key(key)
             enc.start(iv)
@@ -732,7 +782,7 @@ def test():
             dec.start(iv)
             decrypted = dec.finish(ct)
 
-            print("  decrypted %s %d\n" % (hex_encode(decrypted), len(decrypted)))
+            print("  decrypted  %s (%d)\n" % (hex_encode(decrypted), len(decrypted)))
 
 
     def test_mceliece():
@@ -743,9 +793,18 @@ def test():
         mce_ad = 'mce AD'
         mce_ciphertext = mceies_encrypt(mce_pub, rng(), 'ChaCha20Poly1305', mce_plaintext, mce_ad)
 
-        print("mceies len(pt)=%d  len(ct)=%d", (len(mce_plaintext), len(mce_ciphertext)))
+        print("mceies len(pt)=%d  len(ct)=%d" % (len(mce_plaintext), len(mce_ciphertext)))
 
         mce_decrypt = mceies_decrypt(mce_priv, 'ChaCha20Poly1305', mce_ciphertext, mce_ad)
+        print("  mceies plaintext  \'%s\' (%d)" % (mce_plaintext, len(mce_plaintext)))
+        
+        # Since mceies_decrypt() returns bytes in Python3, the following line
+        # needs .decode('utf-8') to convert mce_decrypt from bytes to a
+        # text string (Unicode).
+        # You don't need to add .decode() if
+        # (a) your expected output is bytes rather than a text string, or
+        # (b) you are using Python2 rather than Python3.
+        print("  mceies decrypted  \'%s\' (%d)" % (mce_decrypt.decode('utf-8'), len(mce_decrypt)))
 
         print("mce_pub %s/SHA-1 fingerprint: %s\nEstimated strength %s bits (len %d)\n" %
               (mce_pub.algo_name(), mce_pub.fingerprint("SHA-1"),
@@ -769,9 +828,10 @@ def test():
         sys_rng = rng()
         symkey = sys_rng.get(32)
         ctext = enc.encrypt(symkey, sys_rng)
-        print("ptext  \'%s\'" % hex_encode(symkey))
-        print("ctext   \'%s\'" % hex_encode(ctext))
-        print("decrypt \'%s\'\n" % hex_encode(dec.decrypt(ctext)))
+        print("ptext   \'%s\' (%d)" % (hex_encode(symkey), len(symkey)))
+        print("ctext   \'%s\' (%d)" % (hex_encode(ctext), len(ctext)))
+        print("decrypt \'%s\' (%d)\n" % (hex_encode(dec.decrypt(ctext)),
+                                         len(dec.decrypt(ctext))))
 
         signer = pk_op_sign(rsapriv, 'EMSA4(SHA-384)')
 
@@ -826,24 +886,36 @@ def test():
                   (dh_grp, hex_encode(a_key), hex_encode(b_key)))
 
     def test_certs():
-        cert = x509_cert("src/tests/data/ecc/CSCA.CSCA.csca-germany.1.crt")
+        cert = x509_cert(filename="src/tests/data/ecc/CSCA.CSCA.csca-germany.1.crt")
         print("CSCA (Germany) Certificate\nDetails:")
-        print("SHA-1 fingerprint: %s" % cert.fingerprint("SHA-1"))
-        print("Expected:          32:42:1C:C3:EC:54:D7:E9:43:EC:51:F0:19:23:BD:85:1D:F2:1B:B9")
+        print("SHA-1 fingerprint:   %s" % cert.fingerprint("SHA-1"))
+        print("Expected:            32:42:1C:C3:EC:54:D7:E9:43:EC:51:F0:19:23:BD:85:1D:F2:1B:B9")
 
-        print("Not before:        %s" % cert.time_starts())
-        print("Not after:         %s" % cert.time_expires())
+        print("Not before:          %s" % cert.time_starts())
+        print("Not after:           %s" % cert.time_expires())
 
-        print("Serial number:     %s" % hex_encode(cert.serial_number()))
-        print("Authority Key ID:  %s" % hex_encode(cert.authority_key_id()))
-        print("Subject   Key ID:  %s" % hex_encode(cert.subject_key_id()))
-        print("Public key bits:\n%s\n" % hex_encode(cert.subject_public_key_bits()))
+        print("Serial number:       %s" % hex_encode(cert.serial_number()))
+        print("Authority Key ID:    %s" % hex_encode(cert.authority_key_id()))
+        print("Subject   Key ID:    %s" % hex_encode(cert.subject_key_id()))
+        print("Public key bits:\n%s\n" % b2a_base64(cert.subject_public_key_bits()))
+
+        pubkey = cert.subject_public_key()
+        print("Public key algo:     %s" % pubkey.algo_name())
+        print("Public key strength: %s" % pubkey.estimated_strength() + " bits" )
+
+        dn_fields = ("Name", "Email", "Organization", "Organizational Unit", "Country")
+        for field in dn_fields:
+            try:
+                print("%s: %s" % (field, cert.subject_dn(field, 0)))
+            except Exception:
+                print("Field: %s not found in certificate" % field)
 
         print(cert.to_string())
 
     test_version()
     test_kdf()
     test_pbkdf()
+    test_bcrypt()
     test_hmac()
     test_rng()
     test_hash()
